try:
    from IPython.display import display
    from ipywidgets import HBox, VBox, interactive_output
except ImportError:
    pass
from preliz.internal.plot_helper import get_textboxes
from preliz.ppls.agnostic import (
    inspect_source,
    parse_function_for_pred_textboxes,
    ppl_plot_decorator,
)


def predictive_explorer(
    fmodel,
    samples=50,
    kind_plot="ecdf",
    references=None,
    plot_func=None,
    engine="auto",
    group="prior_predictive",
    var_name=None,
    stats_kwargs=None,
):
    """
    Explore how changing parameters in the prior affects the prior predictive distribution.

    Use this function to interactively explore how a prior predictive distribution changes
    when the priors are changed. It also allows you to visualize how one prior changes when
    another prior is changed, this can be useful for prior that are not set independently, but
    are dependent on each other.

    Parameters
    ----------
    fmodel : callable
        A function with a PreliZ model, a PyMC model, or a Bambi model.
        See examples section below for details
    samples : int, optional
        The number of samples to draw from the prior predictive distribution (default is 50).
    kind_plot : str, optional
        The type of plot to display. Defaults to "kde". Options are "hist" (histogram),
        "kde" (kernel density estimate), "ecdf" (empirical cumulative distribution function).
    references : int, float, list, tuple or dictionary
        Value(s) used as reference points representing prior knowledge. For example expected
        values or values that are considered extreme. Use a dictionary for labeled references.
    plot_func : function
        Custom matplotlib code. Defaults to None. ``kind_plot`` and ``references`` are ignored
        if ``plot_func`` is specified.
    engine : str, optional
        Library used to define the fmodel. Either `preliz`, `pymc` or `bambi`. Default is `auto`.
        The function will automatically select the appropriate library to use based on the fmodel
        provided.
    group : str, optional
        Which group to use. Ignored if the model is defined in `preliz`.
        Defaults to "prior_predictive". You can also pass "prior".
    var_name: str, optional
        The name of the variable to plot. Ignored if the model is defined in `preliz`.
        If "group=prior_predictive" it defaults to the first variable in `observed_RVs`.
        For "prior" it defaults to the last variable in `free_RVs`.
    stats_kwargs : dict, optional
        Additional keyword arguments to pass to the statistics function.
        Defaults to an empty dictionary.
    """
    if stats_kwargs is None:
        stats_kwargs = {}
    source, signature, engine = inspect_source(fmodel)
    model = parse_function_for_pred_textboxes(source, signature, engine)
    textboxes = get_textboxes(signature, model, kind_plot)
    new_fmodel = ppl_plot_decorator(
        fmodel, samples, kind_plot, references, plot_func, engine, group, var_name, stats_kwargs
    )
    out = interactive_output(new_fmodel, textboxes)
    default_names = [
        "__kind__",
        "__set_xlim__",
        "__x_min__",
        "__x_max__",
        "__set_ylim__",
        "__y_min__",
        "__y_max__",
        "__resample__",
    ]
    default_controls = [textboxes[name] for name in default_names]
    params_controls = [v for k, v in textboxes.items() if k not in default_names]

    params_plot = VBox(params_controls + [out])

    display(HBox([params_plot, VBox(default_controls)]))



import numba as nb
import numpy as np
from scipy.special import gammaincc, gammaincinv

from preliz.distributions.distributions import Continuous
from preliz.internal.distribution_helper import all_not_none, eps
from preliz.internal.optimization import optimize_ml
from preliz.internal.special import cdf_bounds, digamma, gammaln, ppf_bounds_cont


class ScaledInverseChiSquared(Continuous):
    r"""
    Scaled Inverse Chi squared  distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \nu, \tau^2) =
                    \frac{(\tau^2\nu/2)^{\nu/2}}{\Gamma(\nu/2)}~
                    \frac{\exp\left[ \frac{-\nu \tau^2}{2 x}\right]}{x^{1+\nu/2}}


    .. plot::
        :context: close-figs


        from preliz import ScaledInverseChiSquared, style
        style.use('preliz-doc')
        nus =  [4., 4, 10.]
        tau2s = [1., 2, 1.]
        for nu, tau2 in zip(nus, tau2s):
            ScaledInverseChiSquared(nu, tau2).plot_pdf(support=(0, 5))

    ========  ===============================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\nu \tau^2 / (\nu - 2)` for :math:`\nu > 2`, else :math:`\infty`
    Variance  :math:`\frac{2 \nu^2 \tau^4}{(\nu - 2)^2 (\nu - 4)}`
              for :math:`\nu > 4`, else :math:`\infty`
    ========  ===============================


    Parameters
    ----------
    nu : float
        Degrees of freedom (nu > 0).
    tau2 : float
        Scale (tau2 > 0).
    """

    def __init__(self, nu=None, tau2=None):
        super().__init__()
        self.nu = nu
        self.tau2 = tau2
        self.support = (0, np.inf)
        self._parametrization(nu, tau2)

    def _parametrization(self, nu=None, tau2=None):
        self.nu = nu
        self.tau2 = tau2
        self.param_names = ("nu", "tau2")
        self.params_support = ((eps, np.inf), (eps, np.inf))
        self.params = (self.nu, self.tau2)
        if all_not_none(nu, tau2):
            self._update(self.nu, self.tau2)

    def _update(self, nu, tau2):
        self.nu = np.float64(nu)
        self.tau2 = np.float64(tau2)
        self.params = (self.nu, self.tau2)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.nu, self.tau2))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.nu, self.tau2)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.nu, self.tau2)

    def logpdf(self, x):
        return nb_logpdf(x, self.nu, self.tau2)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.nu, self.tau2)

    def entropy(self):
        return nb_entropy(self.nu, self.tau2)

    def mean(self):
        return np.where(self.nu > 2, self.nu * self.tau2 / (self.nu - 2), np.inf)

    def mode(self):
        return self.nu * self.tau2 / (self.nu + 2)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return np.where(
            self.nu > 4,
            2 * self.nu**2 * self.tau2**2 / ((self.nu - 2) ** 2 * (self.nu - 4)),
            np.inf,
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return np.where(self.nu > 6, 4 / (self.nu - 6) * np.sqrt(2 * (self.nu - 4)), np.inf)

    def kurtosis(self):
        return np.where(
            self.nu > 8, (12 * (5 * self.nu - 22)) / ((self.nu - 6) * (self.nu - 8)), np.inf
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return (self.nu * self.tau2) / random_state.chisquare(self.nu, size)

    def _fit_moments(self, mean, sigma):
        cv2 = sigma**2 / mean**2
        nu_hat = 4 + 2 / cv2
        tau2_hat = mean * (nu_hat - 2) / nu_hat
        self._update(nu_hat, tau2_hat)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


# @nb.njit(cache=True)
def nb_cdf(x, nu, tau2):
    h_nu = nu / 2
    return cdf_bounds(gammaincc(h_nu, h_nu * tau2 / x), x, 0, np.inf)


# @nb.njit(cache=True)
def nb_ppf(q, nu, tau2):
    h_nu = nu / 2
    vals = h_nu * tau2 / gammaincinv(h_nu, 1 - q)
    return ppf_bounds_cont(vals, q, 0, np.inf)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, nu, tau2):
    if x < 0:
        return -np.inf
    else:
        h_nu = nu / 2
        return (
            -(np.log(x) * (h_nu + 1))
            - (h_nu * tau2) / x
            + np.log(tau2) * h_nu
            - gammaln(h_nu)
            + np.log(h_nu) * h_nu
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, nu, tau2):
    return (-nb_logpdf(x, nu, tau2)).sum()


@nb.njit(cache=True)
def nb_entropy(nu, tau2):
    h_nu = nu / 2
    return h_nu + np.log(h_nu * tau2) + gammaln(h_nu) - (1 + h_nu) * digamma(h_nu)



"""Projective predictive elicitation."""

import warnings

import numpy as np

from preliz.internal.optimization import optimize_pymc_model
from preliz.ppls.agnostic import get_engine
from preliz.ppls.bambi_io import get_pymc_model, write_bambi_string
from preliz.ppls.pymc_io import (
    back_fitting_pymc,
    compile_mllk,
    extract_preliz_distributions,
    get_initial_guess,
    retrieve_variable_info,
    unravel_projection,
    write_pymc_string,
)


def ppe(model, target, engine="auto", new_families=None, random_state=0):
    """
    Prior Predictive Elicitation.

    This method is experimental and under development. It does not offers guarantees of
    correctness. Use with caution and triple-check the results.

    With the projective method we attempt to find a prior that induces
    a prior predictive distribution as close as possible to the target distribution

    Parameters
    ----------
    model : a probabilistic model
        Currently it only works with PyMC model. More PPls coming soon.
    target : a PreliZ distribution or list
        Instance of a PreliZ distribution or a list of tuples where each tuple contains a PreliZ
        distribution and a weight.
        This represents the prior predictive distribution **previously** elicited by the user,
        possibly using other PreliZ's methods to obtain this distribution, such as maxent,
        roulette, quartile, etc.
        This should represent the domain-knowledge of the user and not any observed dataset.
    engine : str
        Library used to define the model. Either `"auto"` (default), `"pymc"` or `"bambi"`.
        Ig `"auto"`, the library is automatically detected.
    new_families : "auto", list or dict
        Defaults to None, the samples are fit to the original prior distribution.
        If "auto", the method evaluates the fit to the original prior plus a set of
        predefined distributions.
        Use a list of PreliZ distribution to specify the alternative distributions
        you want to consider.
        Use a dict with variables names in ``model`` as keys and a list of PreliZ
        distributions as values. This allows to specify alternative distributions
        per variable.
    random_state : {None, int, numpy.random.Generator, numpy.random.RandomState}
        Defaults to 0. Ignored if `method` is `"pathfinder"`.

    Returns
    -------
    new_priors : str
        A string representation of the new priors. The user can copy and paste it into
        the model's code. Ideally, with none to minimal changes.
    """
    warnings.warn(
        """This method is experimental and under development with no guarantees of correctness.
                  Use with caution and triple-check the results."""
    )
    opt_iterations = 400

    rng = np.random.default_rng(random_state)
    engine = get_engine(model) if engine == "auto" else engine

    # Get models information
    if engine == "bambi":
        model = get_pymc_model(model)

    preliz_model = extract_preliz_distributions(model)
    var_info, num_draws = retrieve_variable_info(model)

    # Initial point for optimization
    initial_guess = get_initial_guess(model)
    # compile PyMC model
    fmodel, old_y_value, obs_rvs = compile_mllk(model)
    projection_raveled = optimize_pymc_model(
        fmodel,
        target,
        num_draws,
        opt_iterations,
        initial_guess,
        rng,
    )
    # restore obs_rvs value in the model
    model.rvs_to_values[obs_rvs] = old_y_value

    projection_unraveled = unravel_projection(projection_raveled, var_info, opt_iterations)

    # Backfit `projected_posterior` into the model's prior-families
    projection_backfitted = back_fitting_pymc(
        projection_unraveled, preliz_model, var_info, new_families
    )

    if engine == "bambi":
        new_priors = write_bambi_string(projection_backfitted, var_info)
    elif engine == "pymc":
        new_priors = write_pymc_string(projection_backfitted, var_info)

    return new_priors
